#include <stdio.h>
#include <stm32f4xx.h>
#include <stm32f4xx_rcc.h>
#include <stm32f4xx_gpio.h>
#include <stm32f4xx_tim.h>

#include "misc.h"
#include "mpconfig.h"
#include "qstr.h"
#include "obj.h"
#include "pwm.h"

// PWM
// TIM2 and TIM12 each have CH1on PA5 and PB14
// they are both 32-bit counters
// 16-bit prescaler

void pwm_init(void) {
    // TIM2 & TIM12 clock enable
    RCC_APB1PeriphClockCmd(RCC_APB1Periph_TIM2, ENABLE);
     RCC_APB1PeriphClockCmd(RCC_APB1Periph_TIM12, ENABLE);

	{
		// for PA5
		// GPIOA Configuration: TIM2_CH1 (PA5)
		GPIO_InitTypeDef GPIO_InitStructure;
		GPIO_InitStructure.GPIO_Pin = GPIO_Pin_5;
		GPIO_InitStructure.GPIO_Mode = GPIO_Mode_AF;
		GPIO_InitStructure.GPIO_Speed = GPIO_Speed_25MHz;
		GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
		GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_NOPULL;
		GPIO_Init(GPIOA, &GPIO_InitStructure);

		// Connect TIM2 pins to AF1
		GPIO_PinAFConfig(GPIOA, GPIO_PinSource5, GPIO_AF_TIM2);
	}

	{
		// for PB14
		// GPIOB Configuration: TIM12_CH1 (PB14)
		GPIO_InitTypeDef GPIO_InitStructure;
		GPIO_InitStructure.GPIO_Pin = GPIO_Pin_14;
		GPIO_InitStructure.GPIO_Mode = GPIO_Mode_AF;
		GPIO_InitStructure.GPIO_Speed = GPIO_Speed_25MHz;
		GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
		GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_NOPULL;
		GPIO_Init(GPIOB, &GPIO_InitStructure);

		// Connect TIM2 pins to AF1
		GPIO_PinAFConfig(GPIOB, GPIO_PinSource14, GPIO_AF_TIM12);
	}

	// Compute the prescaler value so TIM2 runs at 10MHz
	uint16_t PrescalerValue = (uint16_t) ((SystemCoreClock / 2) / 10000000) - 1;

	// Time base configuration
	TIM_TimeBaseInitTypeDef TIM_TimeBaseStructure;
	TIM_TimeBaseStructure.TIM_Period = 1023; // timer cycles at 10KHz - 10,000,000 from prescaler / 1000
	TIM_TimeBaseStructure.TIM_Prescaler = PrescalerValue;
	TIM_TimeBaseStructure.TIM_ClockDivision = 0;
	TIM_TimeBaseStructure.TIM_CounterMode = TIM_CounterMode_Up;
	TIM_TimeBaseInit(TIM2, &TIM_TimeBaseStructure);
	TIM_TimeBaseInit(TIM12, &TIM_TimeBaseStructure);

	// PWM Mode configuration
	TIM_OCInitTypeDef TIM_OCInitStructure;
	TIM_OCInitStructure.TIM_OCMode = TIM_OCMode_PWM1;
	TIM_OCInitStructure.TIM_OutputState = TIM_OutputState_Enable;
	TIM_OCInitStructure.TIM_Pulse = 0; // units of 10us, this is the initial duty cycle until duty_cycle is called
	TIM_OCInitStructure.TIM_OCPolarity = TIM_OCPolarity_High;
	TIM_OC1Init(TIM2, &TIM_OCInitStructure); // channel 1
	TIM_OC1Init(TIM12, &TIM_OCInitStructure); // channel 1

	// ?
	TIM_OC1PreloadConfig(TIM2, TIM_OCPreload_Enable); // channel 1
	TIM_OC1PreloadConfig(TIM12, TIM_OCPreload_Enable); // channel 1

	// ?
	TIM_ARRPreloadConfig(TIM2, ENABLE);
	TIM_ARRPreloadConfig(TIM12, ENABLE);

	// TIM2 enable counter
	TIM_Cmd(TIM2, ENABLE);
	TIM_Cmd(TIM12, ENABLE);
}

/******************************************************************************/
/* Micro Python bindings                                                      */

typedef struct _pyb_pwm_obj_t {
    mp_obj_base_t base;
    uint pwm_id;
} pyb_pwm_obj_t;

static void pwm_obj_print(void (*print)(void *env, const char *fmt, ...), void *env, mp_obj_t self_in, mp_print_kind_t kind) {
    pyb_pwm_obj_t *self = self_in;
    print(env, "<Pwm %lu>", self->pwm_id);
}

static mp_obj_t pwm_obj_duty_cycle(mp_obj_t self_in, mp_obj_t duty_cycle) {
    pyb_pwm_obj_t *self = self_in;
    machine_int_t v = mp_obj_get_float(duty_cycle);
    switch (self->pwm_id) {
        case 1: 
			TIM2->CCR1 = v; 
			break;
        case 2: 
			TIM12->CCR1 = v; 
			break;
    }
    return mp_const_none;
}

static MP_DEFINE_CONST_FUN_OBJ_2(pwm_obj_duty_cycle_obj, pwm_obj_duty_cycle);

static const mp_method_t pwm_methods[] = {
    { "duty_cycle", &pwm_obj_duty_cycle_obj },
    { NULL, NULL },
};

static const mp_obj_type_t pwm_obj_type = {
    { &mp_const_type },
    "Pwm",
    .print = pwm_obj_print,
    .methods = pwm_methods,
};

mp_obj_t pyb_Pwm(mp_obj_t pwm_id) {
    pyb_pwm_obj_t *o = m_new_obj(pyb_pwm_obj_t);
    o->base.type = &pwm_obj_type;
    o->pwm_id = mp_obj_get_int(pwm_id);
    return o;
}
